package worker

import (
	"bytes"
	"encoding/json"
	"fmt"
	"log"
	"os"
	"os/signal"
	"syscall"
	"time"

	"github.com/go-redis/redis/v8"
	"github.com/jackc/pgx/v5/pgxpool"
	"github.com/streadway/amqp"

	"git.pbiernat.dev/egommerce/go-api-pkg/consul"
	"git.pbiernat.dev/egommerce/go-api-pkg/fluentd"
	"git.pbiernat.dev/egommerce/go-api-pkg/rabbitmq"

	cnf "git.pbiernat.dev/egommerce/basket-service/internal/config"
	"git.pbiernat.dev/egommerce/basket-service/internal/service"
	"git.pbiernat.dev/egommerce/basket-service/pkg/database"
)

type (
	Worker struct {
		Config   *cnf.Config
		Cache    *redis.Client
		Database *pgxpool.Pool
		Eventbus *amqp.Channel
		Logger   *fluentd.Logger
		Registry *consul.Service
	}
	OptionFn func(*Worker) error // FIXME: similar in server/server.go
)

func New(c *cnf.Config, opts ...OptionFn) *Worker {
	wrk := &Worker{
		Config: c,
	}

	for _, opt := range opts {
		if err := opt(wrk); err != nil {
			log.Fatalf("Failed to attach extension to the worker. Err: %v\n", err)
		}
	}

	return wrk
}

func WithCache(c *cnf.Config) OptionFn {
	return func(w *Worker) error {
		conn := redis.NewClient(&redis.Options{
			Addr:     c.CacheAddr,
			Password: c.CachePassword,
			DB:       0,
		})

		w.Cache = conn

		return nil
	}
}

func WithDatabase(c *cnf.Config) OptionFn {
	return func(w *Worker) error {
		conn, err := database.Connect(c.DbURL)
		if err != nil {
			w.Logger.Log("Failed to connect to Database server: %v\n", err)
			os.Exit(1)
		}

		w.Database = conn

		return nil
	}
}

func WithEventbus(c *cnf.Config) OptionFn {
	return func(w *Worker) error {
		_, chn, err := rabbitmq.Open(c.EventBusURL)
		if err != nil {
			w.Logger.Log("Failed to connect to EventBus server: %v\n", err)
			os.Exit(1)
		}

		err = rabbitmq.NewExchange(chn, c.EventBusExchange)
		if err != nil {
			w.Logger.Log("Failed to declare EventBus exchange: %v\n", err)
			os.Exit(1)
		}

		_, err = chn.QueueDeclare(
			c.EventBusQueue, // name
			false,           // durable
			false,           // delete when unused
			false,           // exclusive
			false,           // no-wait
			nil,             // arguments
		)
		if err != nil {
			w.Logger.Log("Failed to declare EventBus queue: %v\n", err)
			os.Exit(1)
		}

		// w.bindQueues()
		rabbitmq.BindQueueToExchange(chn, c.EventBusQueue, c.EventBusExchange, "catalog.basket.productAddedToBasket")
		rabbitmq.BindQueueToExchange(chn, c.EventBusQueue, c.EventBusExchange, "catalog.basket.productRemovedFromBasket")
		rabbitmq.BindQueueToExchange(chn, c.EventBusQueue, c.EventBusExchange, "catalog.basket.updateQuantity")

		w.Eventbus = chn

		return nil
	}
}

func WithLogger(c *cnf.Config) OptionFn {
	return func(w *Worker) error {
		logHost, logPort, err := fluentd.ParseAddr(c.LoggerAddr)
		if err != nil {
			w.Logger.Log("Failed to parse Fluentd address: %s. Err: %v", c.LoggerAddr, err)
			os.Exit(1)
		}

		logger, err := fluentd.NewLogger(c.Base.GetAppFullName(), logHost, logPort)
		if err != nil {
			w.Logger.Log("Failed to connect to the Fluentd on %s:%d. Err: %v", logHost, logPort, err)
			os.Exit(1)
		}

		w.Logger = logger

		return nil
	}
}

func WithRegistry(c *cnf.Config) OptionFn {
	return func(w *Worker) error {
		registry, err := consul.NewService(c.RegistryAddr, c.Base.AppID, c.Base.AppName, c.Base.AppID, "", "", 0)
		if err != nil {
			w.Logger.Log("Error connecting to %s: %v", c.RegistryAddr, err)
		}

		w.Registry = registry

		go func(w *Worker) { // Fetch Consul KV config and store it in app config
			ticker := time.NewTicker(time.Second * 15)
			for range ticker.C {
				fetchKVConfig(w) // FIXME: duplicated in server
			}
		}(w)

		return nil
	}

}

func fetchKVConfig(w *Worker) { // @FIXME: merge duplication in server.go and worker.go
	config, _, err := w.Registry.KV().Get(w.Config.KVNamespace, nil)
	if err != nil || config == nil {
		return
	}

	kvCnf := bytes.NewBuffer(config.Value)
	decoder := json.NewDecoder(kvCnf)
	if err := decoder.Decode(&w.Config); err != nil {
		return
	}
}

func (w *Worker) Start(while chan struct{}) error {
	go func() {
		sigint := make(chan os.Signal, 1)
		signal.Notify(sigint, os.Interrupt, syscall.SIGINT, syscall.SIGTERM)
		<-sigint

		w.Shutdown()

		close(while)
	}()

	err := w.doWork()
	if err != nil {
		log.Fatalf("Failed to start worker: %s. Reason: %v\n", w.Config.Base.AppID, err)
		close(while)
	}

	w.Logger.Log("Waiting for messages...")

	return nil
}

func (w *Worker) Shutdown() error {
	w.Logger.Log("Worker %s is going down...", w.Config.Base.AppID)

	w.Registry.Unregister()
	w.Eventbus.Close()
	w.Database.Close()
	w.Logger.Log("Gone.")
	w.Logger.Close()

	return nil
}

func (w *Worker) doWork() error {
	msgs, err := w.Eventbus.Consume(
		w.Config.EventBusQueue, // queue
		"",                     // consumer
		false,                  // auto-ack
		false,                  // exclusive
		false,                  // no-local
		false,                  // no-wait
		nil,                    // args
	)
	if err != nil {
		w.Logger.Log("Failed to register a consumer: %s", err)
		os.Exit(1)
	}

	go func() {
		basketSrvc := service.NewBasketService(w.Database, w.Cache, w.Eventbus, w.Logger)

		for m := range msgs {
			go func(m amqp.Delivery) {
				w.processMsg(basketSrvc, m)
			}(m)
		}
	}()

	return nil
}

func (w *Worker) processMsg(srvc *service.BasketService, m amqp.Delivery) {
	msg, err := rabbitmq.Deserialize(m.Body)
	if err != nil {
		w.Logger.Log("Deserialization error: %v\n", err)
		m.Reject(false)

		return
	}

	rnr := &CommandRunner{}

	name := fmt.Sprintf("%s", msg["event"])
	data := (msg["data"]).(map[string]interface{})
	// reqID := (data["request_id"]).(string) // FIXME Check input params!

	w.Logger.Log("Processing message \"%s\" with data: %v\n", name, data)

	var ok = false
	w.Logger.Log("CMD:")
	rnr.cmd = (msg["command"]).(Command)
	w.Logger.Log(": %v", rnr.cmd)
	// case strings.Contains(name, event.EVENT_PRODUCT_ADDED_TO_BASKET):
	// 	basketID := data["basket_id"].(string) // FIXME Check input params!
	// 	productID := data["product_id"]        // FIXME Check input params!

	// 	rnr.cmd = &AddProductToBasketCommand{srvc}
	// 	w.Logger.Log("Adding product #%d to basket #%s. ReqID: #%s", productID, basketID, reqID)
	// case strings.Contains(name, event.EVENT_PRODUCT_REMOVED_FROM_BASKET):
	// 	basketID := data["basket_id"].(string)
	// 	productID := data["product_id"].(float64)

	// 	rnr.cmd = &RemoveProductFromBasketCommand{srvc}
	// 	w.Logger.Log("Removing product #%d from basket #%s. ReqID: #%s", productID, basketID, reqID)
	// }

	ok, _ = rnr.run(data)
	if ok {
		w.Logger.Log("Successful executed message \"%s\"\n", name)
		m.Ack(false)
		return
	}

	w.Logger.Log("Error processing \"%s\": %s (%v)", name, err.Error(), err)
	m.Reject(false) // FIXME: or Nack(repeat until success - maybe message shout know...?
}
